package main;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import utils.BruteForceMethod;
import utils.BruteForceResult;
import utils.BruteForceThread;
import utils.Hash;
import utils.HashingException;
import utils.RadixException;

public class BruteForce {

	/**
	 * Forces a specified hash.
	 * @param args <hash> <maxKeyLength> <method> <threadCount>
	 * @throws Exception if an error occurs.
	 */
	public static void main(String[] args) throws Exception {
		if(args.length < 2){
			printDocumentation();
			System.exit(1);
		}
		
		String hash = args[0];
		int maxLength = Integer.parseInt(args[1]);
		BruteForceMethod method = BruteForceMethod.getInstance((args.length > 2) ? Integer.parseInt(args[2]) : 0);
		int threadCount = (args.length > 3) ? Integer.parseInt(args[3]) : 2;
		
		BruteForce brute = new BruteForce();
		ArrayList<String> list = new ArrayList<String>();
		list.add(hash);
		
		System.out.println("# Brute Force - Hash Function V1.0\n\nEncrypted :");
		for(String s : list)
			System.out.print(s);
		System.out.println("\nMaximum key length: "+maxLength+"\nThread count: "+threadCount+"\nMethod: "+method+"\n\nBruteforcing...");
		
		HashMap<String, String> pairs = new HashMap<String, String>();
		if(threadCount < 2)
			pairs = brute.bruteForce(list, maxLength, method);
		else
			pairs = brute.bruteForceShared(list, method, maxLength, threadCount);
		
		for(String key : pairs.keySet())
			System.out.println(key+" -> "+pairs.get(key));
	}

	private static void printDocumentation() {
		System.out.println("Usage: BruteForce <hash> <maxKeyLength> <method> <threadCount>");
		System.out.println("\t hash: The hashed key.");
		System.out.println("\t maxKeyLength: The maximum length to force.");
		System.out.println("\t method (optional): The set of characters to use: \n\t\t 0->REDUCE\n\t\t 1->COMMON\n\t\t 2->DIGIT_LOWER_UPPER_ACCENT\n\t\t 3->DIGIT_LOWER_UPPER_ACCENT_SPECIAL\n\t\t 4->UNICODE_BASIC\n\t\t 5->UNICODE_BASIC_SUPPLEMENT\n\t\t 6->UNICODE_EXTENDEDA");
		System.out.println("\t threadCount (optional): The number of thread to use");
	}

	/**
	 * Forces the encrypted string with the specified bruteForce method.
	 * @param encrypted The string to force.
	 * @param method The method to use.
	 * @param maxLength The maximum key length to try.
	 * @throws HashingException if error occurs during hashing.
	 */
	public HashMap<String,String> bruteForce(ArrayList<String> encrypted, int maxLength, BruteForceMethod method) throws HashingException {
		BruteForceResult results = new BruteForceResult(encrypted);
		tryKeys(results, 0, (int)Math.pow(method.getCharSet().length, maxLength),method, maxLength);
		
		return results.getResults();
	}
	
	/**
	 * Computes all hashes in the specified range and compares they with the searched encrypted.
	 * @param results The object used to synchronize thread search and which stores results.
	 * @param start The first word to try.
	 * @param end The word that ends the search.
	 * @param method The set of characters used.
	 * @param maxLength The maximum key length to test.
	 * @throws HashingException if an error occurs during hashing.
	 */
	public static void tryKeys(BruteForceResult results, int start, int end, BruteForceMethod method, int maxLength) throws HashingException{
		int current = start;
		String sCurrent = null;
		while(current < end && !results.isFound()){
			sCurrent = intToStringRadix(current, method.getCharSet());
			
			for(int i=0;i<=maxLength-sCurrent.length();++i){
				String sCurrentPad = leftPadding(sCurrent, method.getChar(0), sCurrent.length()+i);
				String currentHash = Hash.hash(sCurrentPad);
				
				if(results.contains(currentHash))
					results.setResult(currentHash, sCurrentPad);
			}
			
			++current;
		}
	}
	
	/**
	 * Forces the encrypted string with the specified bruteForce method, using multiple threads.
	 * @param encrypted The string to force.
	 * @param method The method to use.
	 * @param maxLength The maximum key length to try.
	 * @param threadCount The number of thread to use.
	 * @throws HashingException if error occurs during hashing.
	 * @throws Exception if the number of thread is < 2.
	 */
	public HashMap<String, String> bruteForceShared(ArrayList<String> encrypted, BruteForceMethod method, int maxLength, int threadCount) throws Exception{
		if(threadCount < 2)
			throw new Exception("The number of thread must be >= 2.");
		
		Thread mainThread = Thread.currentThread();
		BruteForceResult results = new BruteForceResult(encrypted, threadCount);
		
		int words = (int) Math.pow(method.getCharSet().length, maxLength);
		int wordsPerThread = words/threadCount;
		int x = 0;
		
		new BruteForceThread(results, x, x+wordsPerThread, method, maxLength, mainThread).start();
		x += wordsPerThread;
		
		for(int t=2;t<threadCount;++t){
			new BruteForceThread(results, x, x+wordsPerThread, method, maxLength, mainThread).start();
			x += wordsPerThread;
		}
		
		new BruteForceThread(results, x, words, method, maxLength, mainThread).start();
		
		while(!results.isAllThreadDone())
			try {
				synchronized (mainThread) {
					mainThread.wait();
				}
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		
		return results.getResults();
	}
	
	/**
	 * Converts the specified integer <i>a</i> from decimal to particular radix (built with characters set).
	 * @param a The integer to convert to particular radix.
	 * @param charSet The base of the radix.
	 * @return A string in radix, built with characters set, equals to the specified integer a.
	 * @throws RadixException if v contains a character which is not included in the radix characters set.
	 */
	public static String intToStringRadix(int a, Character[] charSet){
		StringBuffer res = new StringBuffer();
		int n = charSet.length;
		int q = a, r = 0;
		while(q >= n){
			r = q % n;
			q = q/n;
			res.insert(0, charSet[r]);
		}
		res.insert(0, charSet[q]);
		
		return res.toString();
	}
	
	/**
	 * Converts the specified string <i>v</i> from particular radix (built with characters set) to decimal integer.
	 * @param v The string to convert to decimal integer.
	 * @param charSet The base of the radix.
	 * @return A integer in radix 10 equals to the specified string v.
	 * @throws RadixException if v contains a character which is not included in the radix characters set.
	 */
	public static int stringToIntRadix(String v, Character[] charSet) throws RadixException{
		List<Character> aCharSet = Arrays.asList(charSet);
		int res = 0;
		for(int i=0;i<v.length();++i){
			int index = -1;
			if((index = aCharSet.indexOf(v.charAt(i))) == -1)
				throw new RadixException(v.charAt(i)+" is not a valid character for the specified radix charset.");
				
			res += Math.pow(charSet.length, v.length()-i-1)*index;
		}
		
		return res;
	}
	
	/**
	 * Left pads the specified text with the given character. Total length will be maxLength parameter.
	 * @param text The text to pad.
	 * @param pad The character to add.
	 * @param maxLength The max number of characters.
	 * @return The padded text.
	 */
	public static String leftPadding(String text, char pad, int maxLength) {
		StringBuffer buf = new StringBuffer(maxLength);
	
		for(int i=0;i<maxLength-text.length();++i)
			buf.append(pad);
		
		buf.append(text);
		
		return buf.toString();
	}
	
	/*
	private static Socket connexion(String url){
		Socket sock = null;
		try{
			sock = new Socket (url, 1042);
		}catch (IOException ex){
			return null;
		}
	
		if(sock.isConnected())
			return sock;
		else
			return null;
	}
	
	/*
	protected void bruteForceShared(String encrypted, BruteForceMethod method, int maxLength, int threadCount, String[] ipAddress) throws ClosedChannelException, IOException{
		InetSocketAddress address = new InetSocketAddress("127.0.0.1", 1042);
		ServerSocketChannel serverChannel = ServerSocketChannel.open();
		
		serverChannel.configureBlocking(false);
		serverChannel.socket().bind(address);
		
		Selector selector = Selector.open();
	
		serverChannel.register(selector, SelectionKey.OP_ACCEPT);
		
		for(String ip : ipAddress){
			Socket sock = null;
			if((sock = connexion(ip)) != null)
				System.out.println("Connecté à "+ip);
				try {
					sock.getOutputStream().write("Server2000".getBytes());
				} catch (IOException e) {
					e.printStackTrace();
				}
		}
	
		while(selector.select() > 0) {
			Iterator<SelectionKey> keys = selector.selectedKeys().iterator();
	
			while(keys.hasNext()) {
				SelectionKey key = keys.next();
				keys.remove();
	
				if(key.isValid() && key.isAcceptable()) {
					ServerSocketChannel server = (ServerSocketChannel) key.channel();
					SocketChannel client = server.accept();
					client.configureBlocking(false);
					
					client.register(selector, SelectionKey.OP_READ | SelectionKey.OP_WRITE);
					
				}
				
				if(key.isValid() && key.isReadable()) {
					SocketChannel client = (SocketChannel) key.channel();
					try {
						ByteBuffer buffer = ByteBuffer.allocate(1000);
						
						int bitsLus = client.read(buffer);
						if(bitsLus>0)
							System.out.println("Recu : " + new String(buffer.array(), 0, bitsLus));
						else {
							client.close();
							System.out.println("Fin de connexion");
							key.cancel();
						}
							
					}
					catch(Exception e) {
						e.printStackTrace();
						key.cancel();
						client.close();
					}
				}
				
				if(key.isValid() && key.isWritable()) {
					//écrire éventuellement
				}
			}
		}
		
		selector.close();
		
		serverChannel.close();
	
		#####
		
		Socket echoSocket = null;
	    PrintWriter out = null;
	    BufferedReader in = null;
	
	    try {
	        echoSocket = new Socket("taranis", 7);
	        out = new PrintWriter(echoSocket.getOutputStream(), true);
	        in = new BufferedReader(new InputStreamReader(
	                                    echoSocket.getInputStream()));
	    } catch (UnknownHostException e) {
	        System.err.println("Don't know about host: taranis.");
	        System.exit(1);
	    } catch (IOException e) {
	        System.err.println("Couldn't get I/O for "
	                           + "the connection to: taranis.");
	        System.exit(1);
	    }
	
		BufferedReader stdIn = new BufferedReader(
	                                   new InputStreamReader(System.in));
		String userInput;
	
		while ((userInput = stdIn.readLine()) != null) {
		    out.println(userInput);
		    System.out.println("echo: " + in.readLine());
		}
	
		out.close();
		in.close();
		stdIn.close();
		echoSocket.close();
	}
	*/
}
