package main;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

import utils.Hash;
import utils.HashingException;

public class BruteForceDictionary {
	private final String DICO_MAX4_PATH = "dico4.txt";
	private final String DICO_MAX6_PATH = "dico6.txt";
	private final String DICO_MAX8_PATH = "dico8.txt";
	private final String DICO_PREFIX_NAMES = "names";
	
	private String encrypted;
	private int maxLength;
	private boolean withUCF;
	private boolean withNames;
	private boolean withDigits;
	private BufferedReader br;
	
	public static void main(String[] args) throws HashingException {
		String correctWord;

		if(args.length < 2) {
			System.out.println("Usage: BruteForceDictionary hash maxKeyLength [option...]");
			System.out.println("* hash: The hashed key.");
			System.out.println("* maxKeyLength: The maximum length to force.");
			System.out.println("* option: Additional tests");
			System.out.println("\tnames - Add first names to the search.");
			System.out.println("\tdigits - Complete words to maxKeyLength with digits before and/or after.");
			System.out.println("\tcombine - Try with combinations of words.");
			System.out.println("\tucfirst - Add the first character upper-cased possibilie to the others options.");
			System.out.println("\taccents (default) - Add the unaccented version for each accented word.");
			
			return;
		}
		
		String encrypted = args[0];
		int maxLength = Integer.parseInt(args[1]);

		boolean isUcfirstSet = isOptionSet("ucfirst", args);
		boolean isCombineSet = isOptionSet("combine", args);
		boolean isDigitsSet = isOptionSet("digits", args);
		boolean isNamesSet = isOptionSet("names", args);
		
		BruteForceDictionary bfd = new BruteForceDictionary(encrypted, maxLength, isUcfirstSet, isNamesSet, isDigitsSet);
		
		correctWord = bfd.dicoBruteForceSimple();
		
		if(correctWord == null && isDigitsSet)
			correctWord = bfd.dicoBruteForceDigits();

		if(correctWord == null && isCombineSet)
			correctWord = bfd.dicoBruteForceCombine();
		
		if(correctWord != null)
			System.out.println("Key found: " + correctWord);
		else
			System.out.println("Not found.");
	}
	
	public static boolean isOptionSet(String option, String[] options) {
		for(String s : options) {
			if(s.equalsIgnoreCase(option))
				return true;
		}

		return false;
	}
	
	public BruteForceDictionary(String encrypted, int maxLength, boolean isUcfirstSet, boolean isNamesSet, boolean isDigitsSet) {
		this.withNames = isNamesSet;
		this.withDigits = isDigitsSet;
		this.encrypted = encrypted;
		this.maxLength = maxLength;
		this.withUCF = isUcfirstSet;
	}
	
	/**
	 * Open a (dico) file to read it later.
	 * @param file File to read.
	 */
	private void openFileRead(String file) {
		try {
			InputStream ips = new FileInputStream(file);
			InputStreamReader ipsr = new InputStreamReader(ips);
			br = new BufferedReader(ipsr);
		} catch (FileNotFoundException e) {
			System.err.println(file + " not found.");
		}
	}
	
	/**
	 * Close the current file.
	 */
	private void closeFileRead() {
		try {
			br.close();
		} catch (IOException e) {
			System.err.println("File not readable.");
		}
	}
	
	/**
	 * Pick the next line in the current file.
	 * @return Next line
	 */
	private String getLineFromFileRead() {
		String line = null;
		
		try {
			line = br.readLine();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		return line;
	}
	
	/**
	 * Get the dico file corresponding to a max length.
	 * @param maxLength Max length of the key.
	 * @return The corresponding file. 
	 */
	private String getDicoFile(int maxLength) {
		String prefix = withNames ? DICO_PREFIX_NAMES : "";
		
		switch(maxLength) {
			case 4: return prefix + DICO_MAX4_PATH;
			case 6: return prefix + DICO_MAX6_PATH;
			default: return prefix + DICO_MAX8_PATH;
		}
	}
	
	/**
	 * Load all words smaller or equals than the max length, in a table sort by word length. Used with the ForceBruteCombine.
	 * @param maxLength Max length of the key.
	 * @return Words sorted by length.
	 */
	private ArrayList<ArrayList<String>> loadWordsByLength(int maxLength) {
		String word;
		ArrayList<ArrayList<String>> words = new ArrayList<ArrayList<String>>();
		
		openFileRead(getDicoFile(maxLength));
		
		for(int i = 0; i < maxLength; i++)
			words.add(new ArrayList<String>());
		
		while((word = getLineFromFileRead()) != null)
			words.get(word.length()-1).add(word);

		closeFileRead();
		
		return words;
	}

	private String ucfirst(String word) {
		return Character.toUpperCase(word.charAt(0)) + word.substring(1);
	}
	
	/**
	 * Brute force with combined words.
	 * @return The key found, or null.
	 */
	public String dicoBruteForceCombine() {
		ArrayList<ArrayList<String>> allWords = loadWordsByLength(maxLength);

		try {
			for(ArrayList<String> words : allWords) {
				for(String word : words) {
					String correctWord1 = dicoBruteForceCombineTestKeys(word, maxLength-word.length(), allWords);
					String correctWord2 = withUCF ? dicoBruteForceCombineTestKeys(ucfirst(word), maxLength-word.length(), allWords) : null;
					String correctWord3 = withDigits ? dicoBruteForceDigitsTryKey(word, maxLength-word.length()) : null;

					if(correctWord1 != null)
						return correctWord1;
					
					if(correctWord2 != null)
						return correctWord2;
					
					if(correctWord3 != null)
						return correctWord3;
				}
			}

		} catch (HashingException e) {
			e.printStackTrace();
		}
		
		return null;
	}
	
	/**
	 * Recursive function used with dicoBruteForceCombine.
	 * @param word Current word.
	 * @param toAdd Number of characters remaining to add. 
	 * @param words All words sorted by length.
	 * @return The key found or null.
	 * @throws HashingException if an error occurred during hashing method.
	 */
	public String dicoBruteForceCombineTestKeys(String word, int toAdd, ArrayList<ArrayList<String>> words) throws HashingException {
		
		if(Hash.hash(word).equals(encrypted))
			return word;
		
		if(toAdd < 0)
			return null;
		
		for(int i = 0; i < toAdd; i++) {
			for(String wordAdd : words.get(i)) {
				
				String correctWord1 = dicoBruteForceCombineTestKeys(word + wordAdd, toAdd-(i+1), words);
				String correctWord2 = withUCF ? dicoBruteForceCombineTestKeys(word + ucfirst(wordAdd), toAdd-(i+1), words) : null;
				String correctWord3 = withDigits ? dicoBruteForceDigitsTryKey(word + wordAdd, toAdd-(i+1)) : null;
				String correctWord4 = withDigits && withUCF ? dicoBruteForceDigitsTryKey(word + ucfirst(wordAdd), toAdd-(i+1)) : null;
				
				if(correctWord1 != null)
					return correctWord1;
				
				if(correctWord2 != null)
					return correctWord2;
				
				if(correctWord3 != null)
					return correctWord3;
				
				if(correctWord4 != null)
					return correctWord4;
			}
		}
		
		return null;
	}
	
	/**
	 * Brute force with digits.
	 * @return The key found or null.
	 */
	public String dicoBruteForceDigits() {
		String word = null;

		openFileRead(getDicoFile(maxLength));
		
		try {
			while((word = getLineFromFileRead()) != null) {
				for(int i = 0; i < 10; i++) {
					String correctWord1 = dicoBruteForceDigitsTryKey(word + i, maxLength-word.length()-1);
					String correctWord2 = dicoBruteForceDigitsTryKey(i + word, maxLength-word.length()-1);
					
					String correctWord3 = withUCF ? dicoBruteForceDigitsTryKey(ucfirst(word) + i, maxLength-word.length()-1) : null;
					String correctWord4 = withUCF ? dicoBruteForceDigitsTryKey(i + ucfirst(word), maxLength-word.length()-1) : null;

					if(correctWord1 != null)
						return correctWord1;
					
					if(correctWord2 != null)
						return correctWord2;
					
					if(correctWord3 != null)
						return correctWord3;
					
					if(correctWord4 != null)
						return correctWord4;
				}
			}

		} catch (HashingException e) {
			e.printStackTrace();
		}
		
		closeFileRead();
		
		return null;
	}
	
	/**
	 * Recursive function used with dicoBruteForceDigits  
	 * @param word Current word.
	 * @param toAdd Number of digits remaining to add. 
	 * @return The key found or null.
	 * @throws HashingException if an error occurred during hashing method.
	 */
	private String dicoBruteForceDigitsTryKey(String word, int toAdd) throws HashingException {
	
		if(Hash.hash(word).equals(encrypted))
			return word;
		
		if(toAdd < 0)
			return null;
		
		for(int i = 0; i < 10; i++) {
			String correctWord1 = dicoBruteForceDigitsTryKey(word + i, toAdd-1);
			String correctWord2 = dicoBruteForceDigitsTryKey(i + word, toAdd-1);
			
			if(correctWord1 != null)
				return correctWord1;
			
			if(correctWord2 != null)
				return correctWord2;			
		}
		
		return null;
	}
	
	/**
	 * Simple brute force.
	 * @return The key found or null.
	 */
	public String dicoBruteForceSimple() {
		String word = null;

		openFileRead(getDicoFile(maxLength));
		
		try {
			while((word = getLineFromFileRead()) != null) {
		
				if(Hash.hash(word).equals(encrypted))
					return word;
				
				if(withUCF && Hash.hash(ucfirst(word)).equals(encrypted))
					return ucfirst(word);
			}

		} catch (HashingException e) {
			e.printStackTrace();
		}
		
		closeFileRead();
		
		return null;
	}
}