package utils;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

public class Crypt {
	
	/**
	 * Encrypts <i><b>text</b></i> with <i><b>key</b></i> using method <b>AES/CBC/PKCS5Padding</b>.
	 * @param text The text to encrypt.
	 * @param key The key to use.
	 * @return The encrypted text.
	 * @throws NoSuchPaddingException if <i><b>PKCS5Padding</b></i> is not available from the current installed providers.
	 * @throws NoSuchAlgorithmException if <i><b>AES</b></i> contains a padding scheme that is not available.
	 * @throws IllegalBlockSizeException if <i>AES</i> algorithm is unable to process the input data provided. 
	 * @throws InvalidKeyException if the public key in the given certificate is inappropriate for initializing this cipher, or this cipher is being initialized for decryption or unwrapping keys and requires algorithm parameters that cannot be determined from the public key in the given certificate, or the keysize of the public key in the given certificate has a keysize that exceeds the maximum allowable keysize (as determined by the configured jurisdiction policy files).
	 * @throws InvalidAlgorithmParameterException if <i><b>AES/CBC/PKCS5Padding</b></i> is not available from the current installed providers.
	 */
	public static String encrypt(String text, String key) throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException, InvalidAlgorithmParameterException {
		return encrypt(text, key, EncryptionMethod.AES_CBC_PKCS5Padding);
	}
	
	/**
	 * Decrypts <i><b>text</b></i> with <i><b>key</b></i> using method <b>AES/CBC/PKCS5Padding</b>.
	 * @param text The text to decrypt.
	 * @param key The key to use.
	 * @return The plain text.
	 * @throws NoSuchPaddingException if <i><b>PKCS5Padding</b></i> is not available from the current installed providers.
	 * @throws NoSuchAlgorithmException if <i><b>AES</b></i> contains a padding scheme that is not available.
	 * @throws IllegalBlockSizeException if <i>AES</i> algorithm is unable to process the input data provided. 
	 * @throws InvalidKeyException if the public key in the given certificate is inappropriate for initializing this cipher, or this cipher is being initialized for decryption or unwrapping keys and requires algorithm parameters that cannot be determined from the public key in the given certificate, or the keysize of the public key in the given certificate has a keysize that exceeds the maximum allowable keysize (as determined by the configured jurisdiction policy files).
	 * @throws InvalidAlgorithmParameterException if <i><b>AES/CBC/PKCS5Padding</b></i> is not available from the current installed providers.
	 */
	public static String decrypt(String text, String key) throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException, InvalidAlgorithmParameterException {
		return decrypt(text, key, EncryptionMethod.AES_CBC_PKCS5Padding);
	}
	
	/**
	 * Encrypts <i>text</i> with <i>key</i> using specified <i>method</i>.
	 * @param text The text to encrypt.
	 * @param key The key to use.
	 * @param method The method to use. Method must be an EncryptionMethod object.<br/>
	 * <br/>More details at http://docs.oracle.com/javase/6/docs/technotes/guides/security/StandardNames.html
	 * @return The encrypted text.
	 * @throws NoSuchPaddingException if <i>method</i> is null, empty, in an invalid format, or not available from the current installed providers.
	 * @throws NoSuchAlgorithmException if <i>method</i> contains a padding scheme that is not available.
	 * @throws BadPaddingException if this cipher is in decryption mode, and (un)padding has been requested, but the decrypted data is not bounded by the appropriate padding bytes.
	 * @throws IllegalBlockSizeException if this cipher is a block cipher, no padding has been requested (only in encryption mode), and the total input length of the data processed by this cipher is not a multiple of block size or if this encryption algorithm is unable to process the input data provided. 
	 * @throws InvalidKeyException if the public key in the given certificate is inappropriate for initializing this cipher, or this cipher is being initialized for decryption or unwrapping keys and requires algorithm parameters that cannot be determined from the public key in the given certificate, or the keysize of the public key in the given certificate has a keysize that exceeds the maximum allowable keysize (as determined by the configured jurisdiction policy files).
	 * @throws InvalidAlgorithmParameterException if the given algorithm parameters are inappropriate for this cipher, or this cipher is being initialized for decryption and requires algorithm parameters and params is null, or the given algorithm parameters imply a cryptographic strength that would exceed the legal limits (as determined from the configured jurisdiction policy files).
	 */
	protected static String encrypt(String text, String key, EncryptionMethod method) throws NoSuchAlgorithmException, NoSuchPaddingException, IllegalBlockSizeException, BadPaddingException, InvalidKeyException, InvalidAlgorithmParameterException{
		Cipher c = Cipher.getInstance(method.name().replace('_', '/'));
		
		c.init(Cipher.ENCRYPT_MODE, createKey(key, method), getIV(c.getBlockSize()));
		
		return bytesToHex(c.doFinal(text.getBytes()));
	}
	
	/**
	 * Decrypts <i>text</i> with <i>key</i> using specified <i>method</i>.
	 * @param text The text to decrypt.
	 * @param key The key to use.
	 * @param method The method to use. Method must be an EncryptionMethod object.<br/>
	 * <br/>More details at http://docs.oracle.com/javase/6/docs/technotes/guides/security/StandardNames.html
	 * @return The plain text.
	 * @throws NoSuchPaddingException if <i>method</i> is null, empty, in an invalid format, or not available from the current installed providers.
	 * @throws NoSuchAlgorithmException if <i>method</i> contains a padding scheme that is not available.
	 * @throws BadPaddingException if this cipher is in decryption mode, and (un)padding has been requested, but the decrypted data is not bounded by the appropriate padding bytes.
	 * @throws IllegalBlockSizeException if this cipher is a block cipher, no padding has been requested (only in encryption mode), and the total input length of the data processed by this cipher is not a multiple of block size or if this encryption algorithm is unable to process the input data provided. 
	 * @throws InvalidKeyException if the public key in the given certificate is inappropriate for initializing this cipher, or this cipher is being initialized for decryption or unwrapping keys and requires algorithm parameters that cannot be determined from the public key in the given certificate, or the keysize of the public key in the given certificate has a keysize that exceeds the maximum allowable keysize (as determined by the configured jurisdiction policy files).
	 * @throws InvalidAlgorithmParameterException if the given algorithm parameters are inappropriate for this cipher, or this cipher is being initialized for decryption and requires algorithm parameters and params is null, or the given algorithm parameters imply a cryptographic strength that would exceed the legal limits (as determined from the configured jurisdiction policy files).
	 */
	protected static String decrypt(String crypt, String key, EncryptionMethod method) throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, InvalidAlgorithmParameterException, IllegalBlockSizeException, BadPaddingException{
		Cipher c = Cipher.getInstance(method.name().replace('_', '/'));
		
		c.init(Cipher.DECRYPT_MODE, createKey(key, method), getIV(c.getBlockSize()));
		
		return new String(c.doFinal(hexToBytes(crypt)));
	}
	
	/**
	 * Converts bytes array to hexadecimal string.
	 * @param bytes The array to convert.
	 * @return An hexadecimal string (length equals 2*bytes.length).
	 */
	public static String bytesToHex(byte[] bytes) {
		StringBuilder sb = new StringBuilder(bytes.length*2);
		for (byte b : bytes)
			sb.append( String.format("%02x", b) );
		
		return sb.toString();
	}
	
	/**
	 * Converts an hexadecimal string to a bytes array.
	 * @param hexa The string to convert.
	 * @return An array of bytes (length equals hexa.length()/2).
	 */
	public static byte[] hexToBytes(String hexa) {
	    int len = hexa.length();
	    byte[] bytes = new byte[len / 2];
	    for (int i = 0; i < len; i += 2)
	        bytes[i / 2] = (byte) ((Character.digit(hexa.charAt(i), 16) << 4) + Character.digit(hexa.charAt(i+1), 16));
	    
	    return bytes;
	}
	
	/**
	 * Checks if key has the correct length for the given method and auto truncates/pads if it's not.
	 * @param key The key to check.
	 * @param method The method used for encryption/decryption.
	 * @return A SecretKeySpec object with the appropriate length.
	 */
	protected static SecretKeySpec createKey(String key, EncryptionMethod method) {
		int[] neededLength = method.getKeyLengths();
		byte[] key_b = key.getBytes();
		boolean validKey = false;
		
		for(int i=0;i<neededLength.length;++i)
			if(key_b.length == neededLength[i])
				validKey = true;
		
		if(!validKey)
			key_b = adjustByteTable(key_b, Byte.parseByte("0"), method.getMaxKeyLength());

		return new SecretKeySpec(key_b, method.getMainAlgorithm());
	}
	
	/**
	 * Generates an IV for encryption/decryption method. The IV is a set of zeros. The number of zeros is set by the parameter length.
	 * @param length The number of character for the IV.
	 * @return An IVParameterSpec with the specified length.
	 */
	public static IvParameterSpec getIV(int length){
		return new IvParameterSpec(adjustByteTable(new byte[0],Byte.parseByte("0"),length));	
	}
	
	/**
	 * Adjusts the specified bytes table to the specified length, padding with the pad byte if necessary.
	 * @param tab The byte table to adjust.
	 * @param pad The byte to add if necessary.
	 * @param length The size of the result byte table.
	 * @return An adjusted byte table.
	 */
	public static byte[] adjustByteTable(byte[] tab, byte pad, int length) {
		byte[] res = new byte[length];
		int index = Math.min(length, tab.length);
		
		for(int i=0;i<index;++i)
			res[i] = tab[i];
		
		while(index < length){
			res[index] = pad;
			++index;
		}
	
		return res;
	}
}
